/*
 * gpg-multi.c Copyright (c) 2004 Matthew Wilcox <matthew@wil.cx>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include <sys/mman.h>
#include <sys/types.h>
#include <sys/wait.h>

/*
 * I don't know whether to specify gpg by full path or not.  Both ways have
 * their disadvantages, so "secure by default" and let the user change it.
 */

#define GPG_PATH	"/usr/bin/gpg"

/*
 * If your passphrase is really longer than 1024 bytes, you must
 * be really paranoid (and have hopefully increased the other elements
 * of your security chain to match).
 */

#define PHRASE_MAX_LEN	1024

static char phrase[PHRASE_MAX_LEN];

static void clean_up(char *func)
{
	memset(phrase, 0, PHRASE_MAX_LEN);
	if (func)
		perror(func);
	exit(1);
}

static int get_phrase(char *name)
{
	int count, old_errno;
	struct termios ttyctl;
	tcflag_t old_lflag;

	if (mlock(phrase, PHRASE_MAX_LEN) < 0)
		clean_up("mlock");

	if (!isatty(STDIN_FILENO)) {
		fprintf(stderr, "%s: Input must be from a terminal\n", name);
		exit(1);
	}

	tcgetattr(STDIN_FILENO, &ttyctl);
	old_lflag = ttyctl.c_lflag;
	ttyctl.c_lflag &= ~ECHO;
	tcsetattr(STDIN_FILENO, TCSANOW, &ttyctl);

	printf("Please enter passphrase:");
	fflush(stdout);
	count = read(STDIN_FILENO, phrase, PHRASE_MAX_LEN);

	old_errno = errno;
	ttyctl.c_lflag = old_lflag;
	tcsetattr(STDIN_FILENO, TCSANOW, &ttyctl);
	errno = old_errno;

	if (count < 0)
		clean_up("get_phrase");
	printf("\n");

	return count;
}

static void check_status(int status)
{
	if (WIFEXITED(status)) {
		if (WEXITSTATUS(status) != 0) {
			fprintf(stderr, "Child exited with status %d\n",
					WEXITSTATUS(status));
			clean_up(NULL);
		}
	} else {
		fprintf(stderr, "Child terminated abnormally\n");
		clean_up(NULL);
	}
}

static void do_child(int fd, char *gpg_arg, char *keyid)
{
	char sfd[20];
	sprintf(sfd, "%d", fd);
	execlp(GPG_PATH, "gpg", "--passphrase-fd", sfd, gpg_arg, keyid, NULL);
}

int help(char *name)
{
	printf("%s: Assists manipulating many GnuPG keys\n", name);
	printf("Usage: %s <gpg-operation> <user-id> ...\n", name);

	return 0;
}

int main(int argc, char **argv)
{
	int fds[2];
	int i, len, status;

	if ((argc < 3) || (argv[1][0] != '-') || (argv[1][1] == '?'))
		return help(argv[0]);

	len = get_phrase(argv[0]);

	if (pipe(fds) < 0)
		clean_up("pipe");

	for (i = 2; i < argc; i++) {
		pid_t rc = fork();
		if (rc < 0) {
			clean_up("fork");
		} else if (rc == 0) {
			do_child(fds[0], argv[1], argv[i]);
			clean_up("exec");
		} else {
			if (write(fds[1], phrase, len) != len)
				clean_up("write");
			waitpid(rc, &status, 0);
			check_status(status);
		}
	}

	memset(phrase, 0, PHRASE_MAX_LEN);

	return 0;
}
